import json
from collections import namedtuple


class UnknownMessageError(Exception):
    pass


class MessageType(object):
    # Job status messages
    JOB_FAILED = "JOB_FAILED"
    JOB_STARTED = "JOB_STARTED"
    JOB_UPDATED = "JOB_UPDATED"
    JOB_COMPLETED = "JOB_COMPLETED"
    JOB_CANCELED = "JOB_CANCELED"

    # Job command messages
    START_JOB = "START_JOB"
    CANCEL_JOB = "CANCEL_JOB"


class Message(namedtuple("_Message", ["type", "message"])):
    def serialize(self):
        # check that message type is in one of the message types we define
        assert self.type in (
            t.value for t in list(MessageType)
        ), "Message type not found in predetermined message type list!"

        return json.dumps({"type": self.type, "messsage": self.message})


class SuccessMessage(Message):
    def __new__(cls, job_id, result):
        msg = {'job_id': job_id, 'result': result}
        self = super(SuccessMessage, cls).__new__(cls, type=MessageType.JOB_COMPLETED, message=msg)
        return self


class FailureMessage(Message):
    def __new__(cls, job_id, exc, trace):
        """
        Creates a message indicating an exception was raised when running the job given by job_id.
        The exc is the exception raised, while trace is a string showing the traceback.

        :type trace: str
        :type exc: Exception
        :type job_id: str
        :param job_id: the job_id of the job that failed.
        :param exc: the exception raised during the run of the job.
        :param trace: the string form of the traceback generated by the job.
        :return: a new Message ready to be passed to the messaging backend.
        """

        msg = {'job_id': job_id, 'exception': exc, 'traceback': trace}
        self = super(FailureMessage, cls).__new__(cls, type=MessageType.JOB_FAILED, message=msg)
        return self


class ProgressMessage(Message):
    def __new__(cls, job_id, progress, total_progress, stage=""):
        """
        Creates a Message that updates the progress for the job given by job_id.

        :param job_id: The job_id of the job to update.
        :param progress: the current progress achieved by the running function so far. It should be less than or equal to the value of total_progress.
        :param total_progress: the total amount of progress achievable by the function. This can correspond directly to a concrete action done by the function (such as total number of videos to download, the number of items to process etc.)
        :param stage: an optional argument giving a short name for the current stage the job is on, e.g. 'downloading videos', 'loading subtitles', etc.
        :return: Message

        :type job_id: str
        :type progress: float
        :type total_progress: float
        :type stage: str
        """
        msg = {'job_id': job_id, 'progress': progress, 'total_progress': total_progress, 'stage': stage}
        self = super(ProgressMessage, cls).__new__(cls, type=MessageType.JOB_UPDATED, message=msg)
        return self


class CancelMessage(Message):
    def __new__(cls, job_id):
        """
        Creates a Message that tells the worker(s) to cancel the job with job_id.

        :param job_id: The job_id of the job to cancel.
        :return: Message

        :type job_id: str
        """

        msg = {"job_id": job_id}
        self = super(CancelMessage, cls).__new__(cls, type=MessageType.CANCEL_JOB, message=msg)
        return self


class JobCanceledMessage(Message):
    def __new__(cls, job_id, is_successfully_canceled, last_stage):
        """
        Creates a Message that informs the user that the job has been canceled.

        :param job_id: the job_id
        :param is_successfully_canceled: Whether the job has been canceled voluntarily, or whether it was force-killed.
        :param last_stage: The canceled job can annotate what stage it was at before canceling.
        :return: Message

        :type job_id: str
        :type is_successfully_canceled: bool
        :type last_stage: str
        """

        msg = {"job_id": job_id, "is_successfully_canceled": is_successfully_canceled, "last_stage": last_stage}
        self = super(JobCanceledMessage, cls).__new__(cls, type=MessageType.JOB_CANCELED, message=msg)
        return self



import abc
from collections import defaultdict

from iceqube.exceptions import TimeoutError
from iceqube.async import Event

JOB_EVENT_MAPPING = defaultdict(lambda: Event())


class BaseBackend(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def schedule_job(self, job_details):
        pass

    @abc.abstractmethod
    def mark_job_as_canceled(self, job_id):
        pass

    @abc.abstractmethod
    def get_next_scheduled_job(self):
        pass

    @abc.abstractmethod
    def get_scheduled_jobs(self):
        pass

    @abc.abstractmethod
    def get_job(self, job_id):
        pass

    @abc.abstractmethod
    def complete_job(self, job_id):
        pass

    @abc.abstractmethod
    def mark_job_as_running(self, job_id):
        pass

    @abc.abstractmethod
    def update_job_progress(self, job_id, progress):
        pass

    def wait_for_job_update(self, job_id, timeout=None):
        """
        Blocks until a job given by job_id has updated its state (canceled, completed, progress updated, etc.)
        if timeout is not None, then this function raises iceqube.exceptions.TimeoutError.

        :param job_id: the job's job_id to monitor for changes.
        :param timeout: if None, wait forever for a job update. If given, wait until timeout seconds, and then raise
        iceqube.exceptions.TimeoutError.
        :return: the Job object corresponding to job_id.
        """
        # internally, we register an Event object for each entry in this function.
        # when self.notify_of_job_update() is called, we call Event.set() on all events
        # registered for that job, thereby releasing any threads waiting for that specific job.

        event = JOB_EVENT_MAPPING[job_id]
        event.clear()
        result = event.wait(timeout=timeout)
        job = self.get_job(job_id)

        if result:
            return job
        else:
            raise TimeoutError("Job {} has not received any updates.".format(job_id))

    @staticmethod
    def notify_of_job_update(job_id):
        """
        Unblocks any thread waiting for job updates for the job given by job_id.

        :param job_id: any threads waiting for the job given by job_id is released from waiting.
        :return: None
        """
        JOB_EVENT_MAPPING[job_id].set()



from concurrent.futures import CancelledError


class BaseError(Exception):
    """
    The base exception for all errors raised by iceqube
    """
    pass

class TimeoutError(BaseError):
    """
    An error raised by iceqube.storage.wait() when no job updates come in.
    """


class UserCancelledError(CancelledError):
    """
    An error raised when the user cancels the current job.
    """

    def __init__(self, last_stage):
        self.last_stage = last_stage



import uuid

from iceqube.common.classes import Job, State
from iceqube.messaging.backends import inmem as messaging_inmem
from iceqube.scheduler.classes import Scheduler
from iceqube.storage.backends import inmem as storage_inmem
from iceqube.worker.backends import inmem


class Client(object):
    def __init__(self, app, namespace, **config):
        self.storage = config['storage_backend']
        self.scheduler = config['scheduler']

    def schedule(self, func, *args, **kwargs):
        """
        Schedules a function func for execution.

        The only other special parameter is track_progress. If passed in and not None, the func will be passed in a
        keyword parameter called update_progress:

        def update_progress(progress, total_progress, stage=""):

        The running function can call the update_progress function to notify interested parties of the function's
        current progress.

        All other parameters are directly passed to the function when it starts running.

        :type func: callable or str
        :param func: A callable object that will be scheduled for running.
        :return: a string representing the job_id.
        """

        # if the func is already a job object, just schedule that directly.
        if isinstance(func, Job):
            job = func
        # else, turn it into a job first.
        else:
            job = Job(func, *args, **kwargs)

        job.track_progress = kwargs.pop('track_progress', False)
        job.cancellable = kwargs.pop('cancellable', False)
        job_id = self.storage.schedule_job(job)
        return job_id

    def cancel(self, job_id):
        """
        Mark a job as canceled and remove it from the list of jobs to be executed.
        Send a message to our workers to stop a job.

        :param job_id: the job_id of the Job to cancel.
        """
        self.scheduler.request_job_cancel(job_id)

    def all_jobs(self):
        """
        Return all the jobs scheduled, queued, running, failed or completed.
        Returns: A list of all jobs.

        """
        return self.storage.get_all_jobs()

    def status(self, job_id):
        """
        Returns a Job object corresponding to the job_id. From there, you can query for the following attributes:

        - function string to run
        - its current state (see Job.State for the list of states)
        - progress (returning an int), total_progress (returning an int), and percentage_progress
        (derived from running job.progress/total_progress)
        - the job.exception and job.traceback, if the job's function returned an error

        :param job_id: the job_id to get the Job object for
        :return: the Job object corresponding to the job_id
        """
        return self.storage.get_job(job_id)

    def wait(self, job_id, timeout=None):
        """
        Wait until the job given by job_id has a new update.

        :param job_id: the id of the job to wait for.
        :param timeout: how long to wait for a job state change before timing out.
        :return: Job object corresponding to job_id
        """
        return self.storage.wait_for_job_update(job_id, timeout=timeout)

    def wait_for_completion(self, job_id, timeout=None):
        """
        Wait for the job given by job_id to change to COMPLETED or CANCELED. Raises a
        iceqube.exceptions.TimeoutError if timeout is exceeded before each job change.

        :param job_id: the id of the job to wait for.
        :param timeout: how long to wait for a job state change before timing out.
        """
        while 1:
            job = self.wait(job_id, timeout=timeout)
            if job.state in [State.COMPLETED, State.FAILED, State.CANCELED]:
                return job
            else:
                continue

    def clear(self, force=False):
        """
        Clear all jobs that have succeeded or failed.
        :type force: bool
        :param force: Whether to clear all jobs from the job storage queue, regardless if they've been completed or not.
        """
        self.storage.clear(force=force)


class SimpleClient(Client):
    # types of workers we can spawn
    PROCESS_BASED = inmem.WorkerBackend.PROCESS
    THREAD_BASED = inmem.WorkerBackend.THREAD

    # specify if we want our storage backend to be in-memory
    MEMORY = storage_inmem.StorageBackend.MEMORY

    def __init__(self, app, worker_type=THREAD_BASED, storage_path=MEMORY):
        # simplify configuration by making app and namespace the same thing
        namespace = app

        self.worker_mailbox_name = uuid.uuid4().hex
        self.scheduler_mailbox_name = uuid.uuid4().hex
        self._storage = storage_inmem.StorageBackend(app, app, storage_path)
        self._messaging = messaging_inmem.MessagingBackend()
        self._workers = inmem.WorkerBackend(
            incoming_message_mailbox=self.worker_mailbox_name,
            outgoing_message_mailbox=self.scheduler_mailbox_name,
            msgbackend=self._messaging,
            worker_type=worker_type)
        self._scheduler = Scheduler(
            self._storage,
            self._messaging,
            worker_mailbox=self.worker_mailbox_name,
            incoming_mailbox=self.scheduler_mailbox_name)

        super(SimpleClient, self).__init__(
            app, namespace, storage_backend=self._storage,
            scheduler=self._scheduler)

    def shutdown(self):
        """
        Shutdown the client and all of its managed resources:

        - the workers
        - the scheduler threads

        :return: None
        """
        self._storage.clear()
        self._scheduler.shutdown(wait=False)
        self._workers.shutdown(wait=False)


class InMemClient(SimpleClient):
    """
    A client that starts and runs all jobs in memory. In particular, the following iceqube components are all
    running
    their in-memory counterparts:

    - Scheduler
    - Job storage
    - Workers
    """

    def __init__(self, app, *args, **kwargs):
        super(InMemClient, self).__init__(
            app,
            worker_type=self.THREAD_BASED,
            storage_path=self.MEMORY,
            *args,
            **kwargs)





import logging
import traceback
from abc import ABCMeta, abstractmethod

from iceqube.common.six.moves import queue

from iceqube.common.utils import InfiniteLoopThread
from iceqube.messaging.classes import (
    FailureMessage, MessageType, ProgressMessage, SuccessMessage,
    JobCanceledMessage, )

logger = logging.getLogger(__name__)


class BaseWorkerBackend(object):
    __metaclass__ = ABCMeta

    def __init__(self, incoming_message_mailbox, outgoing_message_mailbox, msgbackend, num_workers=3, *args, **kwargs):
        self.incoming_message_mailbox = incoming_message_mailbox
        self.outgoing_message_mailbox = outgoing_message_mailbox
        self.msgbackend = msgbackend

        self.workers = self.start_workers(num_workers=num_workers)
        self.message_processor = self.start_message_processing()

    @abstractmethod
    def schedule_job(self, job):
        """Manually schedule a job given by job."""
        pass

    @abstractmethod
    def start_workers(self, num_workers):
        pass

    @abstractmethod
    def shutdown_workers(self, wait):
        pass

    @abstractmethod
    def cancel(self, job_id):
        pass

    def shutdown(self, wait=False):
        self.message_processor.stop()
        self.shutdown_workers(wait=wait)

    def start_message_processing(self):
        """
        Starts up the message processor thread, that continuously reads
        messages sent to self.incoming_message_mailbox, and starts or cancels jobs based on the message received.
        Returns: the Thread object.

        """
        t = InfiniteLoopThread(self.process_messages, thread_name="MESSAGEPROCESSOR", wait_between_runs=0.5)
        t.start()
        return t

    def process_messages(self):
        """
        Read from the incoming_message_mailbox and report to the storage backend
        based on the first message found there.
        Returns: None
        """
        try:
            msg = self.msgbackend.pop(self.incoming_message_mailbox)
            self.handle_incoming_message(msg)
        except queue.Empty:
            logger.debug("Worker message queue currently empty.")

    def handle_incoming_message(self, msg):
        """
        Start or cancel a job, based on the msg.

        If msg.type == MessageType.START_JOB, then start the job given by msg.job.

        If msg.type == MessageType.CANCEL_JOB, then try to cancel the job given by msg.job.job_id.

        Args:
            msg (barbequeue.messaging.classes.Message):

        Returns: None

        """
        if msg.type == MessageType.START_JOB:
            job = msg.message['job']
            self.schedule_job(job)
        elif msg.type == MessageType.CANCEL_JOB:
            job_id = msg.message['job_id']
            self.cancel(job_id)

    def report_cancelled(self, job, last_stage):
        msg = JobCanceledMessage(job.job_id, is_successfully_canceled=True, last_stage=last_stage)
        self.msgbackend.send(self.outgoing_message_mailbox, msg)

    def report_success(self, job, result):
        msg = SuccessMessage(job.job_id, result)
        self.msgbackend.send(self.outgoing_message_mailbox, msg)

    def report_error(self, job, exc, trace):
        trace = traceback.format_exc()
        logger.warning("Job {} raised an exception: {}".format(job.job_id, trace))
        msg = FailureMessage(job.job_id, exc, trace)
        self.msgbackend.send(self.outgoing_message_mailbox, msg)

    def update_progress(self, job_id, progress, total_progress, stage=""):
        msg = ProgressMessage(job_id, progress, total_progress, stage)
        self.msgbackend.send(self.outgoing_message_mailbox, msg)


import traceback

from concurrent.futures import CancelledError
from concurrent.futures._base import CANCELLED_AND_NOTIFIED, CANCELLED

from iceqube.exceptions import UserCancelledError
from iceqube.worker.backends.base import BaseWorkerBackend


class WorkerBackend(BaseWorkerBackend):
    # worker types
    PROCESS = 1
    THREAD = 0

    def __init__(self, worker_type=THREAD, *args, **kwargs):
        # Internally, we use conncurrent.future.Future to run and track
        # job executions. We need to keep track of which future maps to which
        # job they were made from, and we use the job_future_mapping dict to do
        # so.
        self.job_future_mapping = {}
        self.future_job_mapping = {}
        self.worker_type = worker_type
        super(WorkerBackend, self).__init__(*args, **kwargs)

        # transfer cancel notification mechanism to something that will work for multiprocess Windows and Android,
        # once we release kolibri 0.5.x
        self.cancel_notifications = {}

    def schedule_job(self, job):
        """
        schedule a job to the type of workers spawned by self.start_workers.


        :param job: the job to schedule for running.
        :return:
        """
        l = _reraise_with_traceback(job.get_lambda_to_execute())

        future = self.workers.submit(l, update_progress_func=self.update_progress, cancel_job_func=self._check_for_cancel)

        # assign the futures to a dict, mapping them to a job
        self.job_future_mapping[future] = job
        self.future_job_mapping[job.job_id] = future

        # callback for when the future is now!
        future.add_done_callback(self.handle_finished_future)
        # add the job to our cancel notifications data structure, with False at first
        self.cancel_notifications[job.job_id] = False

        return future

    def shutdown_workers(self, wait=True):
        self.workers.shutdown(wait=wait)

    def start_workers(self, num_workers):
        if self.worker_type == self.PROCESS:
            from concurrent.futures import ProcessPoolExecutor
            worker_executor = ProcessPoolExecutor
        elif self.worker_type == self.THREAD:
            from concurrent.futures import ThreadPoolExecutor
            worker_executor = ThreadPoolExecutor
        else:
            raise ValueError(
                "WorkerBackend.worker_type must be one of [WorkerBackend.PROCESS, WorkerBackend.THREAD]"
            )

        pool = worker_executor(max_workers=num_workers)
        return pool

    def handle_finished_future(self, future):
        # get back the job assigned to the future
        job = self.job_future_mapping[future]

        try:
            result = future.result()
        except CancelledError as e:
            last_stage = getattr(e, "last_stage", "")
            self.report_cancelled(job, last_stage=last_stage)
            return
        except Exception as e:
            self.report_error(job, e, e.traceback)
            return

        self.report_success(job, result)

    def cancel(self, job_id):
        """
        Request a cancellation from the futures executor pool.
        If that didn't work (because it's already running), then mark
        a special variable inside the future that we can check
        inside a special check_for_cancel function passed to the
        job.
        :param job_id:
        :return:
        """
        future = self.future_job_mapping[job_id]
        is_future_cancelled = future.cancel()

        if is_future_cancelled:  # success!
            return True
        else:
            if future.running():
                # Already running, but let's mark the future as cancelled
                # anyway, to make sure that calling future.result() will raise an error.
                # Our cancelling callback will then check this variable to see its state,
                # and exit if it's cancelled.
                from concurrent.futures._base import CANCELLED
                future._state = CANCELLED
                return False
            else:  # probably finished already, too late to cancel!
                return False

    def _check_for_cancel(self, job_id, current_stage=""):
        """
        Check if a job has been requested to be cancelled. When called, the calling function can
        optionally give the stage it is currently in, so the user has information on where the job
        was before it was cancelled.

        :param job_id: The job_id to check
        :param current_stage: Where the job currently is

        :return: raises a CancelledError if we find out that we were cancelled.
        """

        future = self.future_job_mapping[job_id]
        is_cancelled = future._state in [CANCELLED, CANCELLED_AND_NOTIFIED]

        if is_cancelled:
            raise UserCancelledError(last_stage=current_stage)


def _reraise_with_traceback(f):
    """
    Call the function normally. But if the function raises an error, attach the str(traceback)
    into the function.traceback attribute, then reraise the error.
    Args:
        f: The function to run.

    Returns: A function that wraps f, attaching the traceback if an error occurred.

    """

    def wrap(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception as e:
            traceback_str = traceback.format_exc()
            e.traceback = traceback_str
            raise e

    return wrap



import logging

from iceqube.common.six.moves.queue import Full

from iceqube.common.utils import InfiniteLoopThread
from iceqube.messaging.classes import Message, MessageType, CancelMessage
from iceqube.async import Event


class Scheduler(object):
    def __init__(self, storage_backend, messaging_backend, incoming_mailbox,
                 worker_mailbox):
        self.incoming_mailbox = incoming_mailbox
        self.worker_mailbox = worker_mailbox

        self.storage_backend = storage_backend
        self.scheduler_shutdown_event = Event()

        self.messaging_backend = messaging_backend

        self.scheduler_thread = self.start_scheduler()
        self.worker_message_handler_thread = self.start_worker_message_handler()

    def start_scheduler(self):
        """
        Start the scheduler thread. This thread reads the queue of jobs to be
        scheduled and sends them to the workers.
        Returns: None

        """
        t = InfiniteLoopThread(
            func=self.schedule_next_job,
            thread_name="SCHEDULER",
            wait_between_runs=0.5)
        t.start()
        return t

    def start_worker_message_handler(self):
        """
        Start the worker message handler thread, that loops over messages from workers
        (job progress updates, failures and successes etc.) and then updates the job's status.
        Returns: None

        """
        t = InfiniteLoopThread(
            func=lambda: self.handle_worker_messages(timeout=2),
            thread_name="WORKERMESSAGEHANDLER",
            wait_between_runs=0.5)
        t.start()
        return t

    def shutdown(self, wait=True):
        """
        Shut down the worker message handler and scheduler threads.
        Args:
            wait: If true, block until both threads have successfully shut down. If False, return immediately.

        Returns: None

        """
        self.scheduler_thread.stop()
        self.worker_message_handler_thread.stop()

        if wait:
            self.scheduler_thread.join()
            self.worker_message_handler_thread.join()

    def request_job_cancel(self, job_id):
        """
        Send a message to the workers to cancel the job with job_id. We then mark the job in the storage
        as being canceled.

        :param job_id: the job to cancel
        :return: None
        """
        msg = CancelMessage(job_id)
        self.messaging_backend.send(self.worker_mailbox, msg)
        self.storage_backend.mark_job_as_canceling(job_id)

    def schedule_next_job(self):
        """
        Get the next job in the queue to be scheduled, and send a message
        to the workers to start the job.
        Returns: None

        """
        next_job = self.storage_backend.get_next_scheduled_job()
        # TODO: don't loop over if workers are already all running

        if not next_job:
            logging.debug("No job to schedule right now.")
            return

        try:
            self.messaging_backend.send(self.worker_mailbox,
                                        Message(
                                            type=MessageType.START_JOB,
                                            message={'job': next_job}))
            self.storage_backend.mark_job_as_queued(next_job.job_id)
        except Full:
            logging.debug(
                "Worker queue full; skipping scheduling of job {} for now.".
                    format(next_job.job_id))
            return

    def handle_worker_messages(self, timeout):
        """
        Read messages that are placed in self.incoming_mailbox,
        and then update the job states corresponding to each message.

        Args:
            timeout: How long to wait for an incoming message, if the mailbox is empty right now.

        Returns: None

        """
        msgs = self.messaging_backend.popn(self.incoming_mailbox, n=20)

        for msg in msgs:
            self.handle_single_message(msg)

    def handle_single_message(self, msg):
        """
        Handle one message and modify the job storage appropriately.
        :param msg: the message to handle
        :return: None
        """
        job_id = msg.message['job_id']
        actual_msg = msg.message
        if msg.type == MessageType.JOB_UPDATED:
            progress = actual_msg['progress']
            total_progress = actual_msg['total_progress']
            self.storage_backend.update_job_progress(job_id, progress,
                                                     total_progress)
        elif msg.type == MessageType.JOB_COMPLETED:
            self.storage_backend.complete_job(job_id)
        elif msg.type == MessageType.JOB_FAILED:
            exc = actual_msg['exception']
            trace = actual_msg['traceback']
            self.storage_backend.mark_job_as_failed(job_id, exc, trace)
        elif msg.type == MessageType.JOB_CANCELED:
            self.storage_backend.mark_job_as_canceled(job_id)
        else:
            self.logger.error("Unknown message type: {}".format(msg.type))



import logging
import time
from abc import ABCMeta, abstractmethod
from collections import defaultdict
from iceqube.common.six.moves.queue import Empty, Queue

logger = logging.getLogger(__name__)

INMEM_SUPER_MAILBOX = defaultdict(lambda: Queue())


class BaseMessagingBackend(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def send(self, mailbox, message, **kwargs):
        """
        Send a message to receivers of the mailbox.
        """
        pass

    @abstractmethod
    def wait(self, mailbox, timeout):
        """Wait until there's at least one item in the mailbox.
        Raises Queue.Empty if the mailbox is still empty after the given timeout.

        Note: calling wait() and then pop() is not an atomic operation,
        pop() may still fail if another object has popped in the meantime."""
        pass

    @abstractmethod
    def pop(self, mailbox, timeout=None):
        pass

    @abstractmethod
    def popn(self, mailbox, n=-1):
        """
        Pop at most N number of messages from the mailbox without waiting. If n is less than 0, pop all messages
        it can.
        :type n: int
        :type mailbox: str
        :param mailbox: The mailbox on whose messages to pop from
        :param n: The number of messages to pop. If less than 0, pop all messages.
        :return: The list of messages
        """
        pass

    @abstractmethod
    def count(self, mailbox):
        pass

    @abstractmethod
    def clear(self, mailbox):
        pass


class MessagingBackend(BaseMessagingBackend):
    def __init__(self, *args, **kwargs):
        pass

    def send(self, mailbox, message, *args, **kwargs):
        INMEM_SUPER_MAILBOX[mailbox]._put(message)
        logger.debug("SEND MAILBOX: {} MSG: {}".format(mailbox, message))

    def wait(self, mailbox, timeout=None):
        m = INMEM_SUPER_MAILBOX[mailbox]

        timeout_increment = 0.1

        while True:
            if not m.empty():  # not empty, return now

                return True
            elif timeout <= 0:  # we've gone past our alloted timeout, so raise an error
                raise Empty("Queue currently empty.")
            else:
                time.sleep(timeout_increment)
                timeout -= timeout_increment

    def pop(self, mailbox, timeout=0.2):
        msg = INMEM_SUPER_MAILBOX[mailbox].get(block=True, timeout=timeout)
        logger.debug("POP MAILBOX: {} MSG: {}".format(mailbox, msg))
        return msg

    def popn(self, mailbox, n=-1):
        m = INMEM_SUPER_MAILBOX[mailbox]
        messages = []
        while 1:
            try:
                msg = m.get_nowait()
                messages.append(msg)
            except Empty:
                break
        return messages

    def count(self, mailbox):
        return INMEM_SUPER_MAILBOX[mailbox].qsize()

    def clear(self, mailbox):
        INMEM_SUPER_MAILBOX[mailbox].clear()




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: PythonLibrary(BuildFileAddress(src/BUILD, lib))

from setuptools import setup

setup(**
{   'author': 'Learning Equality',
    'author_email': 'aron@learningequality.org',
    'classifiers': [   'Development Status :: 2 - Pre-Alpha',
                       'Intended Audience :: Developers',
                       'License :: OSI Approved :: MIT License',
                       'Natural Language :: English',
                       'Programming Language :: Python :: 2',
                       'Programming Language :: Python :: 2.7',
                       'Programming Language :: Python :: 3',
                       'Programming Language :: Python :: 3.5',
                       'Programming Language :: Python :: 3.6'],
    'description': '',
    'install_requires': ['SQLAlchemy>=1.1.10', 'futures>=3.1.1'],
    'keywords': ('queue', 'async'),
    'license': 'MIT',
    'long_description': '',
    'name': 'iceqube',
    'package_data': {   },
    'package_dir': {   '': 'src'},
    'packages': [   'iceqube',
                    'iceqube.common',
                    'iceqube.messaging',
                    'iceqube.messaging.backends',
                    'iceqube.scheduler',
                    'iceqube.storage',
                    'iceqube.storage.backends',
                    'iceqube.worker',
                    'iceqube.worker.backends'],
    'url': 'https://github.com/learningequality/iceqube',
    'version': '0.0.5',
    'zip_safe': True}
)
